// OMP-speed-test-programm ((non-)interleaved memory?)

/*
# gcc -Wall -O2 -fopenmp -static -o openmp openmp.c
#    -funroll-loops -ftree-vectorize # -S, omp since gcc-4.2
#    -ftree-vectorizer-verbose=2
*/
//   check asm-code for vectorization (gcc -S)
// GOMP_CPU_AFFINITY="0-15:1"  # works as expected (gomp-1,comp2)
// OMP_PROC_BIND=TRUE # thread bounding, thread affinity (dont work!?)
// OMP_WAIT_POLICY="PASSIVE" OMP_NUM_THREADS=2 ./mem_badram 7168 # 7GB test
// OMP_WAIT_POLICY="ACTIVE"  faster?
// OMP_NUM_THREADS=16 srun -p sw01_short -n 1 -t 00:10:00 ./mem_badram 31000 # /core
// OMP_SCHEDULE=dynamic,1 # default
// OMP_SCHEDULE=static,1  # test, 1M  better for non-interleaved channels
// GOMP_DEBUG=1 # no output seen
// OMP_DISPLAY_ENV=TRUE show OMP-vars
//  better performance for non-interleaved memory (channel=1-way rank=1-way)
// export OMP_SCHEDULE=static,1967568 # 122GiB/8/1024 for faster openmp
//   gives 2.5*ch-performance (4channels) instead default schedule 1.4*ch
//   ToDo check by omp_get_schedule(kind,size)
//
// 2020-06 initial version
//
// 2015-12t100 2x(E5-2630v3-2.4GHz-8c) 2x(2DIMMs/ch 1866MHz*8B=14.9GB/s/ch 4ch=59GB/socket)
// 2020-06t100_16*16GB_r8GB/c_omp: 92.2GB/s 8t  # auto-interleaved 8t? 16t!
// 2020-06t100_16*16GB_r8GB/c_omp: 13.7GB/s 8t  # 1-way/ch 1-way/csrow
// 2020-06jsPC16_2*8GB_r1GB/c:    26.4GB/s  8t 15.7GB/s 1t 78%*34GB/s 2DIMMs*8GB-DDR4-2133MT/s=17GB/s 2ch=34.1GB/s i7-6700-3.4GHz-4c8t w=50%r ab 1-4GB
// 2020-06t100_16*16GB_r15GB/c    95.3GB/s 16t 11.5GB/s 1t # 2cpus/node w=44.2GB/s ab30GB 1cpu8c=59GB/s*81% no-vec
//  OMP_NUM_THREADS=16;GOMP_CPU_AFFINITY="0-15" numactl -l -C 0-15 ./mem_badram $[240*1024] # 30.5GB/s
//  OMP_NUM_THREADS=8;GOMP_CPU_AFFINITY="0-7"   numactl -l -C 0-7  ./mem_badram $[120*1024] # 15.3GB/s
//  OMP_NUM_THREADS=8;GOMP_CPU_AFFINITY="0-7" numactl -m 1 -C 0-7  ./mem_badram $[120*1024] #  3.0GB/s
//  DDR3 1600/1866/2133/2400/2666/2800MHz  t100=1866MHz(4ch*14.9GHz/cpu)
//  DDR4 2133/2400/2666/2800/3000/3200MHz  PC16=2133MHz(2ch*17.1GB/s*78%)
//    latencies 45(2800MHz)-80ns(1600MHz)
//
/*
for x in $(find /sys/devices/system/edac -name \*e_count);do
 echo $x $(cat $x);done
*/      

#include <unistd.h>  /* sleep(60) needs extra h-file */
#include <stdlib.h>  /* system("sleep 600"); */
#include <stdio.h>
#include <sys/time.h>
#include <string.h>   /* strncmp() */
#ifdef _OPENMP
#include <omp.h>
#endif
#include <sys/mman.h>   /* Memory locking functions */

typedef unsigned long int  VULong; // __attribute__((aligned(64)));
//   ULong __attribute__((vector_size(16)));
typedef unsigned long int ULong; // 64bit on 64bit-systems

double fstime(){
 struct timeval tv; if(gettimeofday(&tv, NULL))perror("fstime");
 return tv.tv_sec+1e-6*tv.tv_usec; }

#define VN  4 /* 4*64bit=256bit=avx2 */
#define NUM (1024*1024*(1024/8))  // 1GB
// static needed for openmp to generate vector code !?
//static
// VULong /* volatile */ pa2[NUM]; // 1GB
 
ULong mem_wr(VULong /*volatile*/ * /*const*/ pa,
              long num, VULong lrnd){
// omp for made by hand, using for you not get SIMD, but same speed
#pragma omp parallel
{
  long i, k;
  VULong vlrnd[VN];
  int omp_n=1, omp_i=0;
  ULong i_start=0, i_end=num;
#ifdef _OPENMP
//#pragma omp parallel default(shared)
  omp_i=omp_get_thread_num();
  omp_n=omp_get_num_threads(); // ToDo use i_start + i_end+1
//   __buildin_assume( num % 1024 == 0 ); // gcc 4.9 
  i_start=(omp_i*(num/omp_n));
  i_end=((omp_i+1)*(num/omp_n));
//  printf("omp_rd %2d of %2d %12lu %12lu\n", omp_i, omp_n, i_start, i_end);
#endif
  for (k=0;k<VN;k++) vlrnd[k] = lrnd; // 4*64=256bit avx2
#pragma omp barrier
//    __attribute__ ((aligned (32))) vsum[2]={0,0} ;
// parallel for breaks good vector code in gcc7
// 8*ymm(4*64bit)/loop w12.7GiB/s =  40% *  2*(2133MHz*8B=17GB/s=15.9GiB/s/ch)
//#pragma omp parallel for schedule(static) // private(i,k)
  for (i=i_start; i<i_end; i+=VN) { // howto tell compiler loop is mod 8?
    for (k=0;k<VN;k++) pa[i+k]= vlrnd[k];
  }
#pragma omp barrier
} // omp parallel
  return (ULong)0;
} // mem_wr
/* avx2: .L3:    # twice as _mm_storeu_si128 %xmm0,... (128bit SSE2?)
        vmovdqu %ymm2, (%rbx,%rdi,8)
        vmovdqu %ymm2, 32(%rbx,%rdi,8)
        vmovdqu %ymm2, 64(%rbx,%rdi,8)
        vmovdqu %ymm2, 96(%rbx,%rdi,8)
        vmovdqu %ymm2, 128(%rbx,%rdi,8)
        vmovdqu %ymm2, 160(%rbx,%rdi,8)
        vmovdqu %ymm2, 192(%rbx,%rdi,8)
        vmovdqu %ymm2, 224(%rbx,%rdi,8)
        addq    $32, %rdi
        cmpq    %rdi, %r12
        ja      .L3
*/
/*
gcc -Wall -O2 -funroll-loops -ftree-vectorize -o openmp openmp.c
 -march=native -fopt-info-vec -fopenmp # -S -fopt-info-vec-missed # gcc7.2
*/
// PC16-2ch: 27.4GiB/s=86% w12.9GiB/s  2133MHz*8B=17GB/s=15.9GiB/s/ch
//   lt. intel.com max. 2 channels, max. 34.1GB/s (2133MHz*8B*2ch)
// using OpenMP OMP_NUM_THREADS=4 GOMP_CPU_AFFINITY="0-3" gcc7 openmp
// loop=  0 speed.w:       7.673 GiB/s (1024 MiB) 5GB free
// loop=  0 speed.r:      26.280 GiB/s (1024 MiB)
// loop=  1 speed.w:      13.084 GiB/s (1024 MiB)
// loop=  1 speed.r:      26.148 GiB/s (1024 MiB)
// loop=  1 speed.w:      13.079 GiB/s (2048 MiB) # hand loop 8*ymm/loop
// loop=  1 speed.r:      24.352 GiB/s (2048 MiB)
// t100.n146 86-93GB/s=89.5+-4%, why not const?  2019-01 2s*4ch*14.9GiB/s
// loop=  3 speed.w:      44.170 GiB/s (4096 MiB) 8*ymm/loop=8*avx256
// loop=  3 speed.r:      94.829 GiB/s (4096 MiB) mem_i4:4
// loop=  2 speed.w:      42.867 GiB/s (204800 MiB)
// loop=  2 speed.r:      94.026 GiB/s (204800 MiB) =6channels
// loop=  3 speed.w:      22.181 GiB/s (1024 MiB)  c0-7
// loop=  3 speed.r:      47.052 GiB/s (1024 MiB)
// loop=  2 speed.w:      14.610 GiB/s (204800 MiB) mem_i1:1
// loop=  2 speed.r:      34.277 GiB/s (204800 MiB) =2.3channels
// loop=  2 speed.w:      10.957 GiB/s (102400 MiB) mem_i1:1 c0-7+
// loop=  2 speed.w:      10.909 GiB/s (102400 MiB)        +c8-15  w22GiB/s
// loop=  2 speed.r:      29.887 GiB/s (102400 MiB) c0-7+
// loop=  2 speed.r:      29.589 GiB/s (102400 MiB) +c8-15         r60GiB/s
// loop=  1 speed.w:       8.449 GiB/s (102400 MiB) c0+   = 56% slow-ranks=2*2
// loop=  1 speed.w:       8.391 GiB/s (102400 MiB) +c8
// loop=  1 speed.r:      11.772 GiB/s (102400 MiB) c0+   = 79%
// loop=  1 speed.r:      11.761 GiB/s (102400 MiB) +c8
// openmp serial write parallel read c0-7
// loop=  1 speed.w:       5.580 GiB/s (120000 MiB) (seq=4.2GiB/s)
// loop=  1 speed.r:      32.479 GiB/s (120000 MiB)
//   problem virtual memory mannagement (VMM) + cpu-caching
//  TODO search for fastest core - memory access pattern?
// hpc18: node 0-3 (core0-3+4-7+8-11+12-15) +  node 4-7 (core16-31:4)
//   Epyc-7351 16c 8ch 7001-fam zen-arch (4*4core-chips/socket, each 2ch)
//   16*16GB 2667MT/s=21.3GiB/s/ch (8ch/cpu) = 16GiB/ch 2*4*(32GiB/cnode)
//   mlockall does serial alloc! sometimes faster! (parallel w=50%r)
// loop=  0 speed.w:      10.275 GiB/s (30720 MiB)       # 48% serial-touch
// loop=  0 speed.r:      21.117 GiB/s (30720 MiB) c0 m0 # 99% * 21.1GiB/s/ch
// loop=  1 speed.r:      18.627 GiB/s (30720 MiB) c0 m0 # 87% w=55%r ptouch
// loop=  0 speed.r:      12.452 GiB/s (30720 MiB) c0 m1 (=m2,m3) remoteDie
// loop=  0 speed.r:       6.779 GiB/s (30720 MiB) c0 m4   remoteCPU
// loop=  0 speed.r:       7.511 GiB/s (30720 MiB) c0-3 m4 remoteCPU
// loop=  0 speed.r:      35.945 GiB/s (30720 MiB) c0-3 m0 # 1.69*ch max2ch
// loop=  0 speed.r:      27.524 GiB/s (30720 MiB) c0-7 serial-ini
// loop=  0 speed.r:      71.063 GiB/s (63488 MiB) c0-7 serial-ini
// loop=  0 speed.r:      71.330 GiB/s (30720 MiB) c0-7 parallel-ini
// loop=  0 speed.r:     129.714 GiB/s (129024 MiB) c0-15  serini w=50%r
// loop=  1 speed.r:     142.433 GiB/s (122880 MiB) c0-15  parini w=41%r
// loop=  1 speed.r:     282.198 GiB/s (225280 MiB) c0-31   r=83% w=41%r
//   ToDo max. 225GiB allocatable, why?
//  write-only-low-speed-problem is CPU with cache handling and
//     multicore management, not memory !? write-to-read-ratio=ca.50%
//   cpu cache bypassing? __mm_store_si128 intrinsics?
ULong mem_rd(VULong /*volatile*/ * /*const*/ pa,
              long num, VULong lrnd){
  VULong sum=(VULong)0;
#pragma omp parallel
{
  long i, k;
  VULong vlrnd[VN];
  VULong vsum[VN];
  int omp_n=1, omp_i=0;
  ULong i_start=0, i_end=num;
#ifdef _OPENMP
//#pragma omp parallel default(shared)
  omp_i=omp_get_thread_num();
  omp_n=omp_get_num_threads(); // ToDo use i_start + i_end+1
//   __buildin_assume( num % 1024 == 0 ); // gcc 4.9 
  i_start=(omp_i*(num/omp_n));
  i_end=((omp_i+1)*(num/omp_n));
//  printf("omp_rd %2d of %2d %12lu %12lu\n", omp_i, omp_n, i_start, i_end);
#endif
  for (k=0;k<VN;k++) vlrnd[k] = lrnd; // 4*64=256bit avx2
  for (k=0;k<VN;k++) vsum[k]  = 0;    // 4*64=256bit avx2
#pragma omp barrier
//    __attribute__ ((aligned (32))) vsum[2]={0,0} ;
// parallel for breaks good vector code in gcc7
//#pragma omp parallel for schedule(static) // private(i,k) firstprivate(num,vsum) reduction(|:sum)
  for (i=i_start; i<i_end; i+=VN) { // howto tell compiler loop is mod 8?
    for (k=0;k<VN;k++) 
      vsum[k] |= pa[i+k]^vlrnd[k];// 
      // vsum[k] |= pa[i+k]^pa[i+k+VN]; // 2020-06 does not vectorize, why?
  }
  for (k=0;k<VN;k++)
#pragma omp critical
     sum |= vsum[k];
#pragma omp barrier
} // omp parallel
  if (sum) { printf("error %lu\n", (long unsigned)sum); } // bit error
  return (ULong)sum;
} // mem_rd

int main(int argn, char *argv[]){
  static VULong /* volatile*/ * pa; /* no stack, segfault at getenv */
  ULong rnd=(VULong)0, rnd1, rnd2;
  long ii, i;
  int verbose=0;
  long num=NUM, mb=NUM/1024l/1024*8; // 1024MiB = 1GiB
  double t1, t2; 
  int n_loops=4; /* number of loops (of INVLOOPs with different RND) */
//  int verbose=1; /* 0=quiet(for multinode) ToDo: check slurm_env! */
  int omp_n= 1;  /* openmp-threads */
  int omp_i= 0;  /* openmp -thread-index */
  int sl=sizeof(VULong);  /* sizeLong should be 8 on 64bit or 4 on 32bit */
  // if (argn>1) mb=atoi(argv[1]);    /* memory in MB */
  for (i=1;i<argn;i++) {
    if (argv[i][0]>'0'  && argv[i][0]<='9') mb=atoi(argv[i]); /* MB */ else
    if (argv[i][0]=='-' && argv[i][1]=='l') n_loops=atoi(argv[i]+2); else
    if (argv[i][0]=='-' && argv[i][1]=='v') verbose=atoi(argv[i]+2); else
     { printf(
    "usage: mem_badram [<megabyte=1000>] [-l<nloops=100>] [-v<verbose=0>]\n");
      exit(1); }
  }
  system("date +%F\\ %T"); // 2018-04-14 for better output-documentation
  num= mb*(1024*(1024l/sizeof(VULong)));  /* memory in kB/sizeLong */
  // SLURM_PROCID  #  0..SLURM_NPROCS
  printf("omp_RAM-speed detection 2020.06-2020-06 by JS %luMB l%d v%d\n",
      mb, n_loops, verbose);
#if defined(__GNUC__) // 2020-06 for better documentation
    printf("Compiler: GNUC %d.%d.%d\n",
                  __GNUC__, __GNUC_MINOR__, __GNUC_PATCHLEVEL__);
#endif
#ifdef _OPENMP
#pragma omp parallel default(shared)
  omp_n= omp_get_num_threads();  /* must be called within parallel pragma */
  printf(" using OpenMP OMP_NUM_THREADS= %d\n", omp_n);
  // export OMP_NUM_THREADS=16;GOMP_CPU_AFFINITY="0-31" ./mem_badram 120000
  printf(" using OpenMP GOMP_CPU_AFFINITY=\"%s\"\n",
    getenv("GOMP_CPU_AFFINITY"));
#endif
  printf(" malloc %lu MiB (Mi=2^20) long= %d B = %d bit (arg1)\n",
           mb /* sumMB */, sl /* sizeofLong */, sl*8 /* bits */);
  printf(" malloc %lu MiB = %lu * vlong (Mi=2^20)\n", mb, num);
  fflush(stdout);

//  pa = (ULong *) malloc(mb*1024l*1024); /* alloc mb*MB */
//  pa = (ULong *) aligned_alloc(64, mb*1024l*1024); /* since C11++*/
  posix_memalign((void **)&pa, 64, mb*1024l*1024); // _POSIX_C_SOURCE >= 200112L

  if (!pa) { exit(2); }  // malloc failed
  // touching j+=512 (*8B=4kB) is faster, but gives bad binding on openmp?
  // cat /sys/devices/system/node/node[01]/numastat t100.TODO numa_hit
  //  see t100.node064
//#pragma omp parallel for default(shared) private(j)
//  for (j=0; j<num; j++) pa[j]=0; // single task faster if j+=512 (8B)
  for (ii=0;ii<n_loops;ii++) { // dflt 100 rounds
    rnd1=rand(); // random fill pattern, ToDo better as seed for xorshift-fill
    rnd2=rand();
    rnd=(rnd1<<32) | rnd2;  // 64bit random
#ifdef _OPENMP
//#pragma omp parallel default(shared)
    omp_i=omp_get_thread_num();
    printf(" omp_i=%d / %d\n", omp_i, omp_n);
#endif
    // return number bad blocks, wr+rd
    t1=fstime();
    if (ii==0) mem_wr( pa, num, rnd);
    else       mem_wr( pa, num, rnd);
    t2=fstime();
    if (ii==0) { // must be done after 1st touch, will touch slow serial else
      if ( mlockall(MCL_CURRENT|MCL_FUTURE) == -1 )
        printf("ERROR: mlockall failed (non-root?)\n"); /* root only */ }
    printf(" loop=%3d speed.w:  %10.3f GiB/s (%.0f MiB)\n",
       (int)ii, 1.*mb/(t2-t1)/1024., mb*1.);
    t1=fstime();
               mem_rd( pa, num, rnd);
    t2=fstime(); // double
    printf(" loop=%3d speed.r:  %10.3f GiB/s (%.0f MiB)\n",
       (int)ii, 1.*mb/(t2-t1)/1024., mb*1.);
    fflush(stdout);
  } // loop100
  system("date +%F\\ %T");
  // ToDo: how to integrate that into the jobsystem? extra-shared-queue?
  return 0;
}
